from kabutobashi.attributes.attribute import PageContent


class MetaPage(type):
    """
    値のget, setに関するメタクラス
    """
    def __new__(mcs, name, bases, class_dict):
        for key, value in class_dict.items():
            if isinstance(value, PageContent):
                value.name = key
                value.internal_name = '_' + key
        cls = type.__new__(mcs, name, bases, class_dict)
        return cls


class AbstractPage(object, metaclass=MetaPage):
    pass


class Page(AbstractPage):
    pass



from kabutobashi.crawler.crawler import Crawler
from kabutobashi.crawler.stock_crawler.stock_detail_page import (
    StockBoard, StockDetail
)
from bs4 import BeautifulSoup
from typing import Union


def get_stock_detail(code: Union[str, int]) -> dict:
    """
    単一株の実行日時の詳細情報を取得する関数
    :params code: 取得したい銘柄コード
    :return:
    """
    url = f"https://minkabu.jp/stock/{code}"
    stock_detail_crawler = StockDetailCrawler()
    return stock_detail_crawler(url=url)


class StockDetailCrawler(Crawler):
    """
    インスタンスに付与したurlの株の情報を取得するCrawler
    """

    def __init__(self):
        super().__init__()

    def web_scraping(self, text: str) -> dict:
        """

        :param text:
        :return:
        """
        res = BeautifulSoup(text, 'lxml')
        stock_detail_dict = {}

        stock_board_tag = "ly_col ly_colsize_7 md_box ly_row ly_gutters"
        stock_board = res.find("div", {"class": stock_board_tag})
        # ページ上部の情報を取得
        sb = StockBoard(stock_board)
        stock_detail_dict.update(sb.get_info())

        # ページ中央の情報を取得
        stock_detail = res.find("div", {"class": "stock-detail"})
        sd = StockDetail(stock_detail)
        stock_detail_dict.update(sd.get_info())

        stock_detail_dict['crawl_datetime'] = self.get_crawl_datetime()
        return stock_detail_dict



import pandas as pd
from kabutobashi.method.method import Method
from kabutobashi.attributes.attribute import Field


class ADX(Method):
    """
    以下の指標を計算するクラス
    +DI: 株価の上昇の大きさ
    -DI: 株価の下降の大きさ
    ADX: 株価のトレンドの強さ
    ADXR: ADXの単純移動平均線

    https://www.sevendata.co.jp/shihyou/technical/dmi.html
    https://www.sevendata.co.jp/shihyou/technical/adx.html
    """

    term = Field(required_type=int)
    adx_term = Field(required_type=int)
    adxr_term = Field(required_type=int)

    def __init__(self, term=14, adx_term=14, adxr_term=28):
        super().__init__(method_name="adx")
        self.term = term
        self.adx_term = adx_term
        self.adxr_term = adxr_term

    @staticmethod
    def true_range(x: pd.DataFrame):
        current_high = x['high']
        current_low = x['low']
        prev_close = x['shift_close']
        a = current_high - current_low
        b = current_high - prev_close
        c = prev_close - current_low
        max_ab = max(a, b)
        max_ac = max(a, c)
        return max(max_ab, max_ac)

    @staticmethod
    def compute_dx(x: pd.DataFrame) -> float:
        numerator = abs(x['plus_di'] - x['minus_di'])
        denominator = x['plus_di'] + x['minus_di']
        return numerator / denominator * 100

    @staticmethod
    def fixed_plus_dm(x: pd.DataFrame) -> float:
        if x['plus_dm'] > 0 and x['plus_dm'] > x['minus_dm']:
            return x['plus_dm']
        else:
            return 0

    @staticmethod
    def fixed_minus_dm(x: pd.DataFrame) -> float:
        if x['minus_dm'] > 0 and x['minus_dm'] > x['plus_dm']:
            return x['minus_dm']
        else:
            return 0

    def method(self, _df: pd.DataFrame) -> pd.DataFrame:
        # 利用する値をshift
        _df = _df.assign(
            shift_high=_df['high'].shift(1),
            shift_low=_df['low'].shift(1),
            shift_close=_df['close'].shift(1)
        )
        _df = _df.assign(
            plus_dm=_df.apply(lambda x: x['high'] - x['shift_high'], axis=1),
            minus_dm=_df.apply(lambda x: x['shift_low'] - x['low'], axis=1)
        )
        _df = _df.assign(
            fixed_plus_dm=_df.apply(lambda x: self.fixed_plus_dm(x), axis=1),
            fixed_minus_dm=_df.apply(lambda x: self.fixed_minus_dm(x), axis=1)
        )
        _df = _df.assign(
            true_range=_df.apply(lambda x: self.true_range(x), axis=1),
            sum_tr=lambda x: x['true_range'].rolling(self.term).sum(),
            sum_plus_dm=lambda x: x['fixed_plus_dm'].rolling(self.term).sum(),
            sum_minus_dm=lambda x: x['fixed_minus_dm'].rolling(self.term).sum()
        )

        _df = _df.dropna()
        required_columns = ["sum_plus_dm", "sum_minus_dm", "sum_tr"]
        _df = _df.loc[:, required_columns]
        
        # +DI, -DI
        _df = _df.assign(
            plus_di=_df.apply(lambda x: x['sum_plus_dm'] / x['sum_tr'] * 100, axis=1),
            minus_di=_df.apply(lambda x: x['sum_minus_dm'] / x['sum_tr'] * 100, axis=1)
        )
        _df = _df.assign(
            DX=_df.apply(self.compute_dx, axis=1),
            ADX=lambda x: x['DX'].rolling(self.adx_term).mean(),
            ADXR=lambda x: x['DX'].rolling(self.adxr_term).mean()
        )
        return _df

    @staticmethod
    def _buy_signal(x) -> float:
        """
        DMIとADXを組み合わせた基本パターン
        """
        
        # +DIが-DIを上抜き、ADXが上昇傾向の上向きであれば新規買い
        if x['ADX_trend'] > 0:
            if x['to_plus'] > 0:
                return 1
        
        # +DIが-DIより上に位置している際に、
        # ADXが下向きから上向きに転換した場合
        # if x['diff'] > 0:

    @staticmethod
    def _sell_signal(x) -> float:
        """
        DMIとADXを組み合わせた基本パターン        
        """
        # -DIが+DIを下抜き、ADXが下落傾向の下向きであれば新規空売り
        if x['ADX_trend'] < 0:
            if x['to_minus'] > 0:
                return 1
        
    def signal(self, _df: pd.DataFrame) -> pd.DataFrame:
        """
        buy_signalとsell_signalを付与
        """
        _df['ADX_trend'] = self.trend(_df['ADX'])
        _df['diff'] = _df['plus_di'] - _df['minus_di']
        _df = _df.join(self.cross(_df['diff']))
        
        _df['buy_signal'] = _df.apply(lambda x: self._buy_signal)
        _df['sell_signal'] = _df.apply(lambda x: self._sell_signal)

        return _df



import random


class UserAgent:

    @staticmethod
    def get_user_agent() -> str:
        browser_idx = random.randint(0, 1)
        if browser_idx == 0:
            return Chrome.get_user_agent()
        elif browser_idx == 1:
            return Firefox.get_user_agent()
        else:
            return ""

    @staticmethod
    def get_user_agent_header() -> dict:
        return {"User-Agent": UserAgent.get_user_agent()}


class Chrome:
    mac_chrome_format = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_6) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/{chrome_version} Safari/537.36"
    windows_chrome_format = "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/{chrome_version}"

    os_format_list = [
        mac_chrome_format,
        windows_chrome_format
    ]

    os_size = len(os_format_list) - 1

    version_list = [
        "77.0.3865",
        "78.0.3904",
        "79.0.3945.130",
        "80.0.3987",
        "81.0.4044"
    ]

    version_size = len(version_list) - 1

    @staticmethod
    def get_user_agent() -> str:
        os_idx = random.randint(0, Chrome.os_size)
        chrome_format = Chrome.os_format_list[os_idx]
        version_idx = random.randint(0, Chrome.version_size)
        version = Chrome.version_list[version_idx]
        return chrome_format.format(chrome_version=version)


class Firefox:
    # Firefox
    mac_firefox_format = "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.13; rv:62.0) Gecko/20100101 Firefox/{firefox_version}"
    windows_firefox_format = "Mozilla/5.0 (Windows NT 10.0; Win64; x64; rv:61.0) Gecko/20100101 Firefox/{firefox_version}"

    os_format_list = [
        mac_firefox_format,
        windows_firefox_format
    ]

    os_size = len(os_format_list) - 1

    version_list = [
        "68.4.2",
        "72.0.2",
        "73.0",
        "74.0"
    ]

    version_size = len(version_list) - 1

    @staticmethod
    def get_user_agent() -> str:
        os_idx = random.randint(0, Firefox.os_size)
        chrome_format = Firefox.os_format_list[os_idx]
        version_idx = random.randint(0, Firefox.version_size)
        version = Firefox.version_list[version_idx]
        return chrome_format.format(firefox_version=version)



from kabutobashi.method.method import Method
from kabutobashi.method.sma import SMA
from kabutobashi.method.macd import MACD
from kabutobashi.method.stochastics import Stochastics
from kabutobashi.method.adx import ADX
from kabutobashi.method.bollinger_bands import BollingerBands
from kabutobashi.method.ichimoku import Ichimoku
from kabutobashi.method.momentum import Momentum
from kabutobashi.method.psycho_logical import PsychoLogical



from kabutobashi.attributes.attribute import Field, StockDf
import pandas as pd
import logging


class MetaMethod(type):
    """
    値のget, setに関するメタクラス
    """
    def __new__(mcs, name, bases, class_dict):
        for key, value in class_dict.items():
            if isinstance(value, StockDf):
                value.name = key
                value.internal_name = '_' + key
            elif isinstance(value, Field):
                value.name = key
                value.internal_name = '_' + key
        cls = type.__new__(mcs, name, bases, class_dict)
        return cls


class AbstractMethod(object, metaclass=MetaMethod):
    pass


class Method(AbstractMethod):
    """
    usage
    sma = SMA()
    macd = MCAD()

    # get sma-based-analysis
    sma_df = stock_df.pipe(sma)

    # get macd-based-analysis
    macd_df = stock_df.pipe(macd)
    """
    # 株価を保持するDataFrame
    stock_df = StockDf()

    def __init__(self, method_name: str, *, logger=None):
        """
        :params method_name: 分析手法の名前、__str__()で表示させる文字列
        """
        self.method_name = method_name
        if logger is None:
            self.logger = logging.getLogger()
            self.logger.setLevel(logging.INFO)
        else:
            self.logger = logger

    def __call__(self, stock_df: pd.DataFrame, **kwargs):
        """
        各手法の時系列分析を行い、買いと売りのタイミングを付与
        :params stock_df:
        :params kwargs: {
            "impact": 売りと買いのシグナルを表示させるときに利用,
            "influence": get_impact()にて利用するパラメータ,
            "tail": get_impact()にて利用するパラメータ
        }
        """
        # 各手法指標となる値を計算し、買いと売りの指標を付与
        signal_df = stock_df.pipe(self.validate) \
            .pipe(self.method) \
            .pipe(self.signal)
        # 買い・売りのシグナルを算出する場合
        if "impact" in kwargs:
            return signal_df.pipe(self.get_impact, **kwargs)
        # それ以外は解析結果のdfを返す
        return signal_df

    def __str__(self) -> str:
        """
        分析方法の名前を返す
        """
        return self.method_name

    def validate(self, _df: pd.DataFrame) -> pd.DataFrame:
        self.stock_df = _df
        return self.stock_df

    def method(self, _df: pd.DataFrame) -> pd.DataFrame:
        raise NotImplementedError("please implement your code")

    def signal(self, _df: pd.DataFrame) -> pd.DataFrame:
        raise NotImplementedError("please implement your code")

    @staticmethod
    def cross(
            _s: pd.Series,
            to_plus_name=None,
            to_minus_name=None) -> pd.DataFrame:
        """
        0を基準としてプラスかマイナスのどちらかに振れたかを判断する関数
        :params _s: 対象のpd.Series
        :params to_plus_name: 上抜けた場合のカラムの名前
        :params to_minus_name: 上抜けた場合のカラムの名前
        """
        # shorten vaiable name
        col = "original"
        shifted = "shifted"
        
        # shiftしたDataFrameの作成        
        shift_s = _s.shift(1)
        _df = pd.DataFrame({col: _s, shifted: shift_s})
        
        # 正負が交差した点
        _df = _df.assign(
            is_cross=_df.apply(lambda x: 1 if x[col] * x[shifted] < 0 else 0, axis=1),
            is_higher=_df.apply(lambda x: 1 if x[col] > x[shifted] else 0, axis=1),
            is_lower=_df.apply(lambda x: 1 if x[col] < x[shifted] else 0, axis=1)
        )

        # 上抜けか下抜けかを判断している
        _df = _df.assign(
            to_plus=_df['is_cross'] * _df['is_higher'],
            to_minus=_df['is_cross'] * _df['is_lower']
        )
        if to_plus_name is not None:
            _df = _df.rename(columns={"to_plus": to_plus_name})
        if to_minus_name is not None:
            _df = _df.rename(columns={"to_minus": to_minus_name})
        return _df
        
    @staticmethod
    def trend(_s: pd.Series) -> pd.Series:
        """
        ある系列_sのトレンドを計算する。
        差分のrolling_sumを返す
        """
        # shorten variable name
        col = "original"
        shifted = "shifted"
        
        # shiftしたDataFrameの作成        
        shift_s = _s.shift(1)
        _df = pd.DataFrame({col: _s, shifted: shift_s})
        _df['diff'] = _df['original'] - _df['shifted']
        _df['diff_rolling_sum'] = _df['diff'].rolling(5).sum()
        return _df['diff_rolling_sum']

    @staticmethod
    def get_impact(
            _df: pd.DataFrame,
            influence: int = 2,
            tail: int = 5,
            **kwargs) -> float:
        """
        売りと買いのシグナルの余波の合計値を返す。
        :params _df:
        :params influence:
        :params tail:
        :return: [-1,1]の値をとる。-1: 売り、1: 買いを表す
        """
        _df['buy_impact'] = _df['buy_signal'].ewm(span=influence).mean()
        _df['sell_impact'] = _df['sell_signal'].ewm(span=influence).mean()
        buy_impact_index = _df['buy_impact'].iloc[-tail:].sum()
        sell_impact_index = _df['sell_impact'].iloc[-tail:].sum()
        return round(buy_impact_index - sell_impact_index, 5)



import pytest
import pandas as pd
from kabutobashi.attributes.attribute import (
    Field,
    StockDf,
    PageContent
)
from kabutobashi.errors import StockDfError


# Fieldクラスを検証するためのscript
class MetaAttribute(type):
    """
    値のget, setに関するメタクラス
    """
    def __new__(meta, name, bases, class_dict):
        for key, value in class_dict.items():
            if isinstance(value, Field):
                value.name = key
                value.internal_name = '_' + key
            elif isinstance(value, StockDf):
                value.name = key
                value.internal_name = '_' + key
            elif isinstance(value, PageContent):
                value.name = key
                value.internal_name = '_' + key
        cls = type.__new__(meta, name, bases, class_dict)
        return cls


class AbstractAttribute(object, metaclass=MetaAttribute):
    pass


class AttributeInstance(AbstractAttribute):
    # Field
    param_int = Field(required_type=int)
    param_str = Field(required_type=str)
    param_required = Field(required=True)
    param_candidate = Field(value_candidate=["one", "two"])
    # StockDf
    stock_df = StockDf()
    # PageContent
    page_content = PageContent()


def test_field_instance():
    fi = AttributeInstance()
    # 値をセットする前に取得した場合はNone
    assert fi.param_required is None
    with pytest.raises(ValueError):
        fi.param_required = None
    # Noneの代入は通る
    fi.param_int = None
    with pytest.raises(ValueError):
        fi.param_candidate = "three"
    with pytest.raises(TypeError):
        fi.param_int = "str"
    with pytest.raises(TypeError):
        fi.param_str = 0


def test_stock_df_instance():
    fi = AttributeInstance()
    # 値をセットする前に取得した場合はNone
    assert fi.stock_df is None
    with pytest.raises(StockDfError):
        fi.stock_df = None

    # カンマを含むデータを入れても正しく動くか確認
    data = {
        0: {"open": "1,000.0", "high": "2,000.0", "low": "3,000.0", "close": "4,000.0", "date": "2020-03-01"},
        1: {"open": "1,000.0", "high": "2,000.0", "low": "3,000.0", "close": "4,000.0", "date": "2020-03-02"},
        2: {"open": "1,000.0", "high": "2,000.0", "low": "3,000.0", "close": "4,000.0", "date": "2020-03-03"},
        3: {"open": "1,000.0", "high": "2,000.0", "low": "3,000.0", "close": "4,000.0", "date": "2020-03-04"},
        4: {"open": "1,000.0", "high": "2,000.0", "low": "3,000.0", "close": "4,000.0", "date": "2020-03-05"},
    }
    fi.stock_df = pd.DataFrame.from_dict(data, orient="index")
    columns = fi.stock_df.columns
    assert "open" in columns
    assert "high" in columns
    assert "low" in columns
    assert "close" in columns

    # 日付の項目がdtでも動くかの確認
    data = {
        0: {"open": "1,000.0", "high": "2,000.0", "low": "3,000.0", "close": "4,000.0", "dt": "2020-03-01"},
        1: {"open": "1,000.0", "high": "2,000.0", "low": "3,000.0", "close": "4,000.0", "dt": "2020-03-02"},
        2: {"open": "1,000.0", "high": "2,000.0", "low": "3,000.0", "close": "4,000.0", "dt": "2020-03-03"},
        3: {"open": "1,000.0", "high": "2,000.0", "low": "3,000.0", "close": "4,000.0", "dt": "2020-03-04"},
        4: {"open": "1,000.0", "high": "2,000.0", "low": "3,000.0", "close": "4,000.0", "dt": "2020-03-05"},
    }
    fi.stock_df = pd.DataFrame.from_dict(data, orient="index")
    columns = fi.stock_df.columns
    assert "open" in columns
    assert "high" in columns
    assert "low" in columns
    assert "close" in columns

    # date列がない場合にErrorを返す
    data = {
        0: {"open": "1,000.0", "high": "2,000.0", "low": "3,000.0", "close": "4,000.0"},
        1: {"open": "1,000.0", "high": "2,000.0", "low": "3,000.0", "close": "4,000.0"},
        2: {"open": "1,000.0", "high": "2,000.0", "low": "3,000.0", "close": "4,000.0"},
        3: {"open": "1,000.0", "high": "2,000.0", "low": "3,000.0", "close": "4,000.0"},
        4: {"open": "1,000.0", "high": "2,000.0", "low": "3,000.0", "close": "4,000.0"},
    }
    with pytest.raises(StockDfError):
        fi.stock_df = pd.DataFrame.from_dict(data, orient="index")

    # 日付の項目がdtでも動くかの確認
    data = {
        0: {"open": "1.0", "high": "2.0", "low": "3.0", "close": "4.0", "dt": "2020-03-01", "date": "2020-03-01"},
        1: {"open": "1.0", "high": "2.0", "low": "3.0", "close": "4.0", "dt": "2020-03-02", "date": "2020-03-02"},
        2: {"open": "1.0", "high": "2.0", "low": "3.0", "close": "4.0", "dt": "2020-03-03", "date": "2020-03-03"},
        3: {"open": "1.0", "high": "2.0", "low": "3.0", "close": "4.0", "dt": "2020-03-04", "date": "2020-03-04"},
        4: {"open": "1.0", "high": "2.0", "low": "3.0", "close": "4.0", "dt": "2020-03-05", "date": "2020-03-05"},
    }
    with pytest.raises(StockDfError):
        fi.stock_df = pd.DataFrame.from_dict(data, orient="index")

    # valueにfloatに変換できな文字列が含まれている場合にエラーを返すテスト
    data = {
        0: {"open": "-------", "high": "2,000.0", "low": "3,000.0", "close": "4,000.0", "date": "2020-03-01"},
        1: {"open": "1,000.0", "high": "2,000.0", "low": "3,000.0", "close": "4,000.0", "date": "2020-03-02"},
        2: {"open": "1,000.0", "high": "2,000.0", "low": "3,000.0", "close": "4,000.0", "date": "2020-03-03"},
        3: {"open": "1,000.0", "high": "2,000.0", "low": "3,000.0", "close": "4,000.0", "date": "2020-03-04"},
        4: {"open": "1,000.0", "high": "2,000.0", "low": "3,000.0", "close": "4,000.0", "date": "2020-03-05"},
    }
    with pytest.raises(StockDfError):
        fi.stock_df = pd.DataFrame.from_dict(data, orient="index")


def test_page_content_instance():
    fi = AttributeInstance()
    # 値をセットする前に取得した場合はNone
    assert fi.page_content is None
    with pytest.raises(ValueError):
        fi.page_content = None



from kabutobashi.crawler.crawler import Crawler
from bs4 import BeautifulSoup
from typing import Union


def get_ipo_list_from_year(year: Union[str, int]) -> dict:
    """
    IPOのリストを取得する関数
    """
    base_url = "https://96ut.com/ipo/list.php"
    year_str = None
    if type(year) is int:
        year_str = str(year)
    else:
        year_str = year

    if year_str == "2020":
        # 実行年の場合
        url = base_url
    else:
        # 過去年の場合
        url = f"{base_url}?year={year}"
    ipo_list_crawler = IPOListCrawler()
    return ipo_list_crawler(url=url)


class IPOListCrawler(Crawler):
    """
    指定した年にIPOした企業名と銘柄コードを取得する
    """

    def __init__(self):
        super().__init__()

    def web_scraping(self, text: str) -> dict:
        """
        :params text: webページ
        """
        res = BeautifulSoup(text, 'lxml')
        table_content = res.find("div", {"class": "tablewrap"})
        table_thead = table_content.find("thead")
        # headの取得
        table_head_list = []
        for th in table_thead.find_all("th"):
            table_head_list.append(th.get_text())

        # bodyの取得
        table_tbody = table_content.find("tbody")
        whole_result = {}
        for idx, tr in enumerate(table_tbody.find_all("tr")):
            table_body_dict = {}
            for header, td in zip(table_head_list, tr.find_all("td")):
                table_body_dict[header] = td.get_text().replace("\n", "")
            whole_result[idx] = table_body_dict
        return whole_result



import pytest
import kabutobashi as ps


def test_crawl_page_not_found():
    page = "https://minkabu.jp/stock/994"
    with pytest.raises(ps.errors.CrawlPageNotFoundError):
        ps.get_web_page(page)


def test_crawl_page_detail():
    result = ps.get_stock_detail(4395)
    assert result is not None
    assert type(result) is dict


def test_crawl_ipo_list():
    result = ps.get_ipo_list_from_year(2019)
    assert result is not None
    assert type(result) is dict


def test_crawl_week_52_high_low_list():
    result = ps.get_52_weeks_high_low("high")
    assert result is not None
    assert type(result) is dict



import pandas as pd
from kabutobashi.method.method import Method
from kabutobashi.attributes.attribute import Field


class Ichimoku(Method):
    """
    https://kabu.com/investment/guide/technical/04.html
    """

    short_term = Field()
    medium_term = Field()
    long_term = Field

    def __init__(self, short_term=12, medium_term=26, long_term=52):
        super().__init__(method_name="ichimoku")
        self.short_term = short_term
        self.medium_term = medium_term
        self.long_term = long_term

    def method(self, _df: pd.DataFrame) -> pd.DataFrame:
        _df = _df.assign(
            # 短期の線
            short_max=lambda x: x['close'].rolling(self.short_term).max(),
            short_min=lambda x: x['close'].rolling(self.short_term).min(),
            # 中期の線
            medium_max=lambda x: x['close'].rolling(self.medium_term).max(),
            medium_min=lambda x: x['close'].rolling(self.medium_term).min(),
            # 長期線
            long_max=lambda x: x['close'].rolling(self.long_term).max(),
            long_min=lambda x: x['close'].rolling(self.long_term).min()
        )

        # 指標の計算
        _df = _df.assign(
            line_change=lambda x: (x['short_max'] + x['short_min']) / 2,
            line_base=lambda x: (x['medium_max'] + x['medium_min']) / 2,
            # 先行線
            proceding_span_1=lambda x: (x['line_change'] + x['line_base']) / 2,
            proceding_span_2=lambda x: (x['long_max'] + x['long_min']) / 2
        )

        # 値のshift
        _df = _df.assign(
            proceding_span_1=_df['proceding_span_1'].shift(26),
            proceding_span_2=_df['proceding_span_2'].shift(26),
            delayed_span=_df['close'].shift(26)
        )
        return _df

    def signal(self, _df: pd.DataFrame) -> pd.DataFrame:
        return _df



from kabutobashi.crawler.crawler import Crawler

# 単一銘柄の詳細情報を取得するクラス
from kabutobashi.crawler.stock_crawler.stock_detail_crawler import (
    get_stock_detail
)

from kabutobashi.crawler.stock_crawler.ipo_list_crawler import (
    get_ipo_list_from_year
)

from kabutobashi.crawler.stock_crawler.weeks_52_high_low_crawler import (
    get_52_weeks_high_low
)


def get_web_page(url: str) -> str:
    crawler = Crawler()
    return crawler.get_url_text(url)



import pandas as pd
from kabutobashi.method.method import Method
from kabutobashi.attributes.attribute import Field


class PsychoLogical(Method):
    """
    https://www.sevendata.co.jp/shihyou/technical/psycho.html
    """
    upper_threshold = Field(required_type=float)
    lower_threshold = Field(required_type=float)
    psycho_term = Field(required_type=int)

    def __init__(
            self,
            upper_threshold=0.75,
            lower_threshold=0.25,
            psycho_term=12):
        super().__init__(method_name="psycho_logical")
        self.upper_threshold = upper_threshold
        self.lower_threshold = lower_threshold
        self.psycho_term = psycho_term

    def method(self, _df: pd.DataFrame) -> pd.DataFrame:
        _df['shift_close'] = _df['close'].shift(1)
        _df['diff'] = _df.apply(lambda x: x['close']-x['shift_close'], axis=1)
        
        _df['is_raise'] = _df['diff'].apply(lambda x: 1 if x > 0 else 0)
        
        _df['psycho_sum'] = _df['is_raise'].rolling(self.psycho_term).sum()
        _df['psycho_line'] = _df['psycho_sum'].apply(lambda x: x/self.psycho_term)
        
        _df['bought_too_much'] = _df['psycho_line'].apply(lambda x: 1 if x > self.upper_threshold else 0)
        _df['sold_too_much'] = _df['psycho_line'].apply(lambda x: 1 if x < self.lower_threshold else 0)
        return _df

    def signal(self, _df: pd.DataFrame) -> pd.DataFrame:
        _df['buy_signal'] = _df['sold_too_much']
        _df['sell_signal'] = _df['bought_too_much']
        return _df



class Field(object):
    """
    様々な値の保持に用いるクラス。
    値を代入する際にバリデーションが行われる。
    """
    def __init__(
            self,
            required=False,
            required_type=None,
            value_candidate=None):
        """
        :params required: True if value is required
        :params required_type: specify type if the type is fixed
        :params value_candidte: list the candidate values
        """
        self.name = None
        self.internal_name = None
        self.required: bool = required
        self.required_type: type = required_type
        self.value_candidate: list = value_candidate

    def __get__(self, instance, instance_type):
        return getattr(instance, self.internal_name, None)

    def __set__(self, instance, value):
        if value is None:
            if self.required:
                raise ValueError(f"The field is required and none is invalid")
            else:
                return

        if self.required_type is not None:
            if type(value) is not self.required_type:
                raise TypeError(f"{self.required_type} is required, but {type(value)}(:={value}) is given")
        if self.value_candidate is not None:
            if value not in self.value_candidate:
                raise ValueError(f"{value} is not accepted, candidates are [{','.join(self.value_candidate)}]")
        setattr(instance, self.internal_name, value)



import pandas as pd
from kabutobashi.method.method import Method
from kabutobashi.attributes.attribute import Field


class MACD(Method):

    short_term = Field()
    long_term = Field()
    macd_span = Field()

    def __init__(
            self,
            short_term: int = 12,
            long_term: int = 26,
            macd_span: int = 9):
        super().__init__(method_name="macd")
        self.short_term = short_term
        self.long_term = long_term
        self.macd_span = macd_span

    def method(self, _df: pd.DataFrame) -> pd.DataFrame:
        """
        macdを基準として今後上昇するかどうかをスコアで返す。
        値が大きければその傾向が高いことを表している。
        最小値は0で、最大値は無限大である。
        :param _df:
        :return:
        """
        # histogramが図として表現されるMACDの値
        _df = _df.assign(
            # MACDの計算
            ema_short=lambda x: x['close'].ewm(span=self.short_term).mean(),
            ema_long=lambda x: x['close'].ewm(span=self.long_term).mean(),
            macd=lambda x: x['ema_short'] - x['ema_long'],
            signal=lambda x: x['macd'].ewm(span=self.macd_span).mean(),
            # ヒストグラム値
            histogram=lambda x: x['macd'] - x['signal'],
        )
        return _df

    def signal(self, _df: pd.DataFrame) -> pd.DataFrame:
        # 正負が交差した点
        _df = _df.join(self.cross(_df['histogram']))
        _df = _df.rename(columns={
            "to_plus": "buy_signal",
            "to_minus": "sell_signal"
        })
        return _df

    # def visualize(self, _df: pd.DataFrame):
    #     fig, (ax1, ax2) = plt.subplots(nrows=2, ncols=1, gridspec_kw={'height_ratios': [3, 1]}, figsize=(6, 5))
    #     # x軸のオートフォーマット
    #     fig.autofmt_xdate()
    #
    #     # set candlestick
    #     self.add_ax_candlestick(ax1, _df)
    #
    #     # plot macd
    #     ax2.plot(_df.index, _df['macd'], label="macd")
    #     ax2.plot(_df.index, _df['signal'], label="signal")
    #     ax2.bar(_df.index, _df['histogram'], label="histogram")
    #     ax2.legend(loc="center left")  # 各線のラベルを表示
    #
    #     ax1.legend(loc="best")  # 各線のラベルを表示
    #     return fig



from kabutobashi.crawler.page import Page
from kabutobashi.attributes.attribute import PageContent
from bs4 import BeautifulSoup


class Week52HighLowStockPricePageTable(Page):
    """
    を取得する
    """
    def __init__(self, content: BeautifulSoup):
        # テーブル要素をすべて取得
        table = content.find_all("tr")
        self.result_list = list()
        for t in table:
            row = Week52HighLowStockPricePageRow(t)
            self.result_list.append(row.get_info())

    def get_info(self) -> dict:
        return {"data": self.result_list}


class Week52HighLowStockPricePageRow(Page):
    # ここからcrawlするページのタグ
    cell = "tv-screener-table__cell"
    signal = "tv-screener-table__signal"
    close_tag = f"tv-data-table__cell {cell} {cell}--big"
    # タグ
    volatility_up_tag = f"{close_tag} {cell}--up"
    volatility_down_tag = f"{close_tag} {cell}--down"
    buy_tag = f"{signal} {signal}--buy"
    strong_buy_tag = f"{signal} {signal}--strong-buy"
    sell_tag = f"{signal} {signal}--sell"
    strong_sell_tag = f"{signal} {signal}--strong-sell"
    # 取得する内容
    code = PageContent(tag1="a", required=True)
    brand_name = PageContent(tag1="span", required=True)
    close = PageContent(tag1="td", _class1=close_tag, required=True)
    volatility_up = PageContent(tag1="td", _class1=volatility_up_tag)
    volatility_down = PageContent(tag1="td", _class1=volatility_down_tag)
    buy = PageContent(tag1="span", _class1=buy_tag, alternative_data="")
    strong_buy = PageContent(tag1="span", _class1=strong_buy_tag, alternative_data="")
    sell = PageContent(tag1="span", _class1=sell_tag, alternative_data="")
    strong_sell = PageContent(tag1="span", _class1=strong_sell_tag, alternative_data="")

    def __init__(self, row: BeautifulSoup):
        self.code = row
        self.brand_name = row
        self.close = row
        self.buy = row
        self.strong_buy = row
        self.sell = row
        self.strong_sell = row
        self.buy_or_sell = f"{self.buy}{self.strong_buy}{self.sell}{self.strong_sell}"
        volatility_dict = self._crawl_volatility_info(row)
        self.volatility_ratio = volatility_dict.get("volatility_ratio", "-")
        self.volatility_value = volatility_dict.get("volatility_value", "-")

    def get_info(self) -> dict:
        return {
            "code": self.code,
            "brand_name": self.brand_name,
            "close": self.close,
            "buy_or_sell": self.buy_or_sell,
            "volatility_ratio": self.volatility_ratio,
            "volatility_value": self.volatility_value
        }

    def _crawl_volatility_info(self, table: BeautifulSoup) -> dict:
        # 上がったか、下がったかを判断するリスト
        evaluate_list = list()
        evaluate_list.extend(table.find_all("td", class_=self.volatility_up_tag))
        evaluate_list.extend(table.find_all("td", class_=self.volatility_down_tag))
        return self._volatility_replace_method(volatility_list=evaluate_list)

    @staticmethod
    def _volatility_replace_method(volatility_list: list) -> dict:
        """
        volatility_listにはweb pageから取得したデータが入っている
        suffixには高値か底値かを表す接頭辞を表す
        """
        if len(volatility_list) == 0:
            return {}
        rate_candidate = volatility_list[0].text
        value_candidate = volatility_list[1].text
        return {
            "volatility_ratio": rate_candidate,
            "volatility_value": value_candidate
            }



# methods to analysis
from kabutobashi.method.api import (
    Method,
    SMA,
    MACD,
    Stochastics,
    ADX,
    BollingerBands,
    Ichimoku,
    Momentum,
    PsychoLogical
)

from kabutobashi.core import (
    # technical analysis function
    analysis_with,
    # get buy or sell signal value
    get_impact_with
)

# functions to load or save files
from kabutobashi.io.api import (
    # read csv data
    read_csv,
    # provide example stock data
    example_data,
    # read stock data
    read_stock_csv
)

# import errors
from kabutobashi import errors

# classes or functions about crawl web pages
from kabutobashi.crawler.api import (
    # beautifulsoupを利用してウェブページを取得する
    get_web_page,
    # 単一の株価の詳細情報を取得する
    get_stock_detail,
    # ある年にIPOした銘柄の情報を取得する
    get_ipo_list_from_year,
    # 52週高値・底値を取得する関数
    get_52_weeks_high_low
)

from .utilities import (
    # n営業日前までの日付のリストを返す関数
    get_past_n_days
)

# create and initialize instance
sma = SMA(short_term=5, medium_term=21, long_term=70)
macd = MACD(short_term=12, long_term=26, macd_span=9)
stochastics = Stochastics()
adx = ADX()
bollinger_bands = BollingerBands()
ichimoku = Ichimoku()
momentum = Momentum()
psycho_logical = PsychoLogical()


# comparable tuple
VERSION = (0, 1, 0)
# generate __version__ via VERSION tuple
__version__ = ".".join(map(str, VERSION))

# module level doc-string
__doc__ = """
kabutobashi
=======

**kabutobashi** is a Python package to analysis stock data with measure
analysis methods, such as MACD, SMA, etc.

Main Features
-------------
Here are the things that kabutobashi does well:
 - Easy crawl.
 - Easy analysis.
"""



from kabutobashi.crawler.page import Page
from kabutobashi.attributes.attribute import PageContent
from bs4 import BeautifulSoup


class StockBoard(Page):
    """
    ページ上部の
    stock_label: {code} {市場}
    name: {銘柄名}
    date: 株価MM/DD
    close: 終値
    を取得する
    """
    stock_label = PageContent(tag1="div", _class1="stock_label")
    name = PageContent(tag1="p", _class1="md_stockBoard_stockName")
    date = PageContent(tag1="h2", _class1="stock_label fsl")
    close = PageContent(tag1="div", _class1="stock_price")

    def __init__(self, stock_board: BeautifulSoup):
        super().__init__()
        self.stock_label = stock_board
        self.name = stock_board
        self.date = stock_board
        self.close = stock_board

    def get_info(self) -> dict:
        return {
            "stock_label": self.stock_label,
            "name": self.name,
            "close": self.close,
            "date": self.date,
        }


class StockDetail(Page):
    """
    ページ中央の情報を取得する
    """

    industry_type = PageContent(
        tag1="div",
        _class1="ly_content_wrapper size_ss")

    def __init__(self, stock_detail: BeautifulSoup):
        self.industry_type = stock_detail

        info = {}
        for li in stock_detail.find_all("li", {"class": "ly_vamd"}):
            info[li.find("dt").get_text()] = li.find("dd").get_text()
        self.open = info['始値']
        self.high = info['高値']
        self.low = info['安値']
        self.unit = info['単元株数']
        self.per = info['PER(調整後)']
        self.psr = info['PSR']
        self.pbr = info['PBR']
        self.volume = info['出来高']
        self.market_capitalization = info['時価総額']
        self.issued_shares = info['発行済株数']

    def get_info(self) -> dict:
        return {
            "industry_type": self.industry_type,
            "open": self.open,
            "high": self.high,
            "low": self.low,
            "unit": self.unit,
            "per": self.per,
            "psr": self.psr,
            "pbr": self.pbr,
            "volume": self.volume,
            "market_capitalization": self.market_capitalization,
            "issued_shares": self.issued_shares
        }



import pytest
import kabutobashi as ps


@pytest.fixture(scope="module", autouse=True)
def var_stock_df():
    stock_df = ps.example_data()
    yield stock_df


def test_example_data(var_stock_df):
    columns = var_stock_df.columns
    assert "date" in columns
    assert "open" in columns
    assert "close" in columns
    assert "high" in columns
    assert "low" in columns


def test_read_stock_data():
    none_paths = []
    df = ps.read_stock_csv(none_paths)
    assert df is None

    paths = [
        "../data/dt=2020-02-29.csv",
        "../data/dt=2020-03-01.csv"
        ]
    df = ps.read_stock_csv(paths)
    columns = df.columns
    assert "code" in columns
    assert "close" in columns
    assert "open" in columns
    assert "high" in columns
    assert "low" in columns
    assert "per" in columns
    assert "pbr" in columns
    assert "psr" in columns
    assert "unit" in columns
    assert "volume" in columns
    assert "dt" in columns
    assert "market" in columns


def test_analysis_with_sma(var_stock_df):
    analysis_df = ps.analysis_with(var_stock_df, ps.sma)
    columns = analysis_df.columns
    assert "sma_short" in columns
    assert "sma_medium" in columns
    assert "sma_long" in columns
    assert "buy_signal" in columns
    assert "sell_signal" in columns


def test_analysis_with_macd(var_stock_df):
    analysis_df = ps.analysis_with(var_stock_df, ps.macd)
    columns = analysis_df.columns
    assert "ema_short" in columns
    assert "ema_long" in columns
    assert "signal" in columns
    assert "macd" in columns
    assert "histogram" in columns
    assert "buy_signal" in columns
    assert "sell_signal" in columns


def test_analysis_with_stochastics(var_stock_df):
    analysis_df = ps.analysis_with(var_stock_df, ps.stochastics)
    columns = analysis_df.columns
    assert "K" in columns
    assert "D" in columns
    assert "SD" in columns
    assert "buy_signal" in columns
    assert "sell_signal" in columns


def test_analysis_with_adx(var_stock_df):
    analysis_df = ps.analysis_with(var_stock_df, ps.adx)
    columns = analysis_df.columns
    assert "plus_di" in columns
    assert "minus_di" in columns
    assert "DX" in columns
    assert "ADX" in columns
    assert "ADXR" in columns
    assert "buy_signal" in columns
    assert "sell_signal" in columns


def test_analysis_with_ichimoku(var_stock_df):
    analysis_df = ps.analysis_with(var_stock_df, ps.ichimoku)
    columns = analysis_df.columns
    assert "line_change" in columns
    assert "line_base" in columns
    assert "proceding_span_1" in columns
    assert "proceding_span_2" in columns
    assert "delayed_span" in columns


def test_analysis_with_momentum(var_stock_df):
    analysis_df = ps.analysis_with(var_stock_df, ps.momentum)
    columns = analysis_df.columns
    assert "momentum" in columns
    assert "sma_momentum" in columns
    assert "buy_signal" in columns
    assert "sell_signal" in columns


def test_analysis_with_spycho_logical(var_stock_df):
    analysis_df = ps.analysis_with(var_stock_df, ps.psycho_logical)
    columns = analysis_df.columns
    assert "psycho_line" in columns
    assert "bought_too_much" in columns
    assert "sold_too_much" in columns
    assert "buy_signal" in columns
    assert "sell_signal" in columns


def test_analysis_with_bollinger_bands(var_stock_df):
    analysis_df = ps.analysis_with(var_stock_df, ps.bollinger_bands)
    columns = analysis_df.columns
    assert "upper_2_sigma" in columns
    assert "lower_2_sigma" in columns
    assert "over_upper_continuity" in columns
    assert "over_lower_continuity" in columns
    assert "buy_signal" in columns
    assert "sell_signal" in columns


def test_get_impact_with(var_stock_df):
    result_1 = ps.get_impact_with(var_stock_df, ps.sma)
    assert "sma" in result_1.keys()
    result_2 = ps.get_impact_with(var_stock_df, [ps.sma, ps.macd])
    assert "sma" in result_2.keys()
    assert "macd" in result_2.keys()


def test_io_read_csv():
    _df = ps.read_csv(1)
    assert _df is None
    _df = ps.read_csv("../data/stooq.csv")
    assert _df is not None



import math
import pandas as pd
from kabutobashi.method.method import Method


class Stochastics(Method):
    def __init__(self):
        super().__init__(method_name="stochastics")

    def method(self, _df: pd.DataFrame) -> pd.DataFrame:
        """
        買いのシグナルを計算で求める
        ・%K・%D共に20％以下の時に、%Kが%Dを下から上抜いた時
        ・%D・スロー%D共に20％以下の時に、%Dがスロー%Dを下から上抜いた時
        https://www.moneypartners.co.jp/support/tech/sct.html
        :param _df:
        :return:
        """

        _df['close'] = _df['close'].astype(float)
        _df['low'] = _df['low'].astype(float)
        _df['high'] = _df['high'].astype(float)
        _df['K'] = Stochastics.fast_stochastic_k(_df['close'], _df['low'], _df['high'], 9)
        _df['D'] = Stochastics.fast_stochastic_d(_df['K'])
        _df['SD'] = Stochastics.slow_stochastic_d(_df['D'])
        return _df

    def signal(self, _df: pd.DataFrame) -> pd.DataFrame:
        """
        買いと売りに関する指標を算出する
        """
        _df = _df.assign(
            shift_K=lambda x: x['K'].shift(1),
            shift_D=lambda x: x['D'].shift(1),
            shift_SD=lambda x: x['SD'].shift(1),
        )

        # 複数引数は関数を利用することで吸収
        _df['buy_signal'] = _df.apply(self._buy_signal_index, axis=1)
        _df['sell_signal'] = _df.apply(self._sell_signal_index, axis=1)
        return _df

    @staticmethod
    def fast_stochastic_k(close, low, high, n):
        return ((close - low.rolling(window=n, center=False).min()) / (
                    high.rolling(window=n, center=False).max() - low.rolling(window=n, center=False).min())) * 100

    @staticmethod
    def fast_stochastic_d(stochastic_k):
        # ストキャスの%Dを計算（%Kの3日SMA）
        return stochastic_k.rolling(window=3, center=False).mean()

    @staticmethod
    def slow_stochastic_d(stochastic_d):
        # ストキャスの%SDを計算（%Dの3日SMA）
        return stochastic_d.rolling(window=3, center=False).mean()

    @staticmethod
    def _buy_signal_index(x: pd.Series) -> float:
        return Stochastics.buy_signal_index(x['K'], x['D'], x['SD'], x['shift_K'], x['shift_D'], x['shift_SD'])

    @staticmethod
    def buy_signal_index(current_k, current_d, current_sd, prev_k, prev_d, prev_sd) -> float:
        if (current_k > 30) | (current_d > 30) | (current_sd > 30):
            return 0

        # %K・%D共に20％以下の時に、%Kが%Dを下から上抜いた時
        if current_k < 20 and current_d < 20:
            if (prev_d > prev_k) and (current_d < current_k):
                return current_k - current_d

        # %D・スロー%D共に20％以下の時に、%Dがスロー%Dを下から上抜いた時
        if current_d < 20 and current_sd < 20:
            if (prev_sd > prev_d) and (current_sd < current_d):
                return current_d - current_sd
        return 1 / math.exp(math.pow(current_k - 20, 2) / 100
                            + math.pow(current_d - 20, 2) / 100
                            + math.pow(current_sd - 20, 2) / 100)

    @staticmethod
    def _sell_signal_index(x: pd.Series) -> float:
        return Stochastics.sell_signal_index(x['K'], x['D'], x['SD'], x['shift_K'], x['shift_D'], x['shift_SD'])

    @staticmethod
    def sell_signal_index(current_k, current_d, current_sd, prev_k, prev_d, prev_sd) -> float:
        if (current_k < 70) | (current_d < 70) | (current_sd < 70):
            return 0

        # %K・%D共に80％以上の時に、%Kが%Dを上から下抜いた時
        if current_k > 80 and current_d > 80:
            if (prev_d < prev_k) and (current_d > current_k):
                return current_d - current_k

        # %D・スロー%D共に80％以上の時に、%Dがスロー%Dを上から下抜いた時
        # %D・スロー%D共に20％以下の時に、%Dがスロー%Dを下から上抜いた時
        if current_d > 80 and current_sd > 80:
            if (prev_sd < prev_d) and (current_sd > current_d):
                return current_d - current_sd
        return 1 / math.exp(math.pow(current_k - 20, 2) / 100
                            + math.pow(current_d - 20, 2) / 100
                            + math.pow(current_sd - 20, 2) / 100)



import pandas as pd
from kabutobashi.method.method import Method
from kabutobashi.attributes.attribute import Field


class BollingerBands(Method):
    """
    https://www.sevendata.co.jp/shihyou/technical/bori.html
    """
    band_term = Field()
    continuity_term = Field()

    def __init__(self, band_term=12, continuity_term=10):
        super().__init__(method_name="bollinger_bands")
        self.band_term = band_term
        self.continuity_term = continuity_term

    def method(self, _df: pd.DataFrame) -> pd.DataFrame:
        _df = _df.assign(
            mean=_df['close'].rolling(self.band_term).mean(),
            std=_df['close'].rolling(self.band_term).std()
        )
        _df = _df.assign(
            upper_2_sigma=_df.apply(lambda x: x['mean'] + x['std'] * 2, axis=1),
            lower_2_sigma=_df.apply(lambda x: x['mean'] - x['std'] * 2, axis=1)
        )
        return _df

    def signal(self, _df: pd.DataFrame) -> pd.DataFrame:
        _df = _df.assign(
            over_upper=_df.apply(lambda x: 1 if x['close'] > x['upper_2_sigma'] else 0, axis=1),
            over_lower=_df.apply(lambda x: 1 if x['close'] < x['lower_2_sigma'] else 0, axis=1),
            over_upper_continuity=lambda x: x['over_upper'].rolling(self.continuity_term).sum(),
            over_lower_continuity=lambda x: x['over_lower'].rolling(self.continuity_term).sum()
        )
        
        _df['buy_signal'] = _df['over_upper'].apply(lambda x: 1 if x > 0 else 0)
        _df['sell_signal'] = _df['over_lower'].apply(lambda x: 1 if x > 0 else 0)
        return _df



from datetime import datetime, timedelta
import jpholiday


def get_past_n_days(current_date: str, n: int = 60) -> list:
    """
    土日と祝日を考慮したn営業日前までの日付のリストを返す関数
    :param current_date: n日前を計算する起点となる日
    :param n: n日前
    :return: ["%Y-%m-%d", "%Y-%m-%d", "%Y-%m-%d", ...]
    """
    end_date = datetime.strptime(current_date, "%Y-%m-%d")
    # 2倍しているのは土日や祝日が排除されるため
    back_n_days = n * 2
    date_candidate = [end_date - timedelta(days=d) for d in range(back_n_days)]
    # 土日を除く
    filter_weekend = [d for d in date_candidate if d.weekday() < 5]
    # 祝日を除く
    filter_holiday = [d for d in filter_weekend if not jpholiday.is_holiday(d)]
    # 文字列に日付を変えてreturn
    return list(map(lambda x: x.strftime("%Y-%m-%d"), filter_holiday[:n]))





class PyStockBaseError(Exception):
    pass


class PyStockCrawlerError(PyStockBaseError):
    """
    Crawlが失敗したときに返すエラー
    """

    def __init__(self, url: str = ""):
        self.url = url

    def __str__(self):
        return f"error occurred when crawling [{self.url}]"


class CrawlPageNotFoundError(PyStockCrawlerError):
    """
    crawlしたいページがない場合に返すエラー
    """

    def __init__(self, url: str = ""):
        super().__init__(url=url)

    def __str__(self):
        return f"Page not found [{self.url}]"


class TagNotFoundError(PyStockCrawlerError):
    """
    crawlしたいページに対象のtagがない場合に返すエラー
    """

    def __init__(self, tag):
        super().__init__(url="")
        self.tag = tag

    def __str__(self):
        return f"tag [{self.tag}] not found"


class StockDfError(PyStockBaseError):
    pass


class PyStockMethodError(PyStockBaseError):
    pass


class PyStockVisualizeError(PyStockBaseError):
    pass


class PyStockAttributeError(PyStockBaseError):
    pass



from kabutobashi.attributes.attribute import Field
from kabutobashi.crawler.user_agent import UserAgent
from kabutobashi.errors import (
    CrawlPageNotFoundError,
    PyStockCrawlerError
)
from datetime import datetime, timedelta, timezone
import requests


class MetaCrawler(type):
    """
    値のget, setに関するメタクラス
    """
    def __new__(mcs, name, bases, class_dict):
        for key, value in class_dict.items():
            if isinstance(value, Field):
                value.name = key
                value.internal_name = '_' + key
        cls = type.__new__(mcs, name, bases, class_dict)
        return cls


class AbstractCrawler(object, metaclass=MetaCrawler):
    pass


class Crawler(AbstractCrawler):

    def __init__(self):
        """
        インスタンスを生成
        """
        pass

    def __call__(self, **kwargs) -> dict:
        url = None
        text = None
        if "url" in kwargs:
            url = kwargs['url']
            text = self.get_url_text(target_url=url)
            url = None
        if "text" in kwargs:
            text = kwargs['text']
        # 両方に値が含まれている場合は例外を投げる
        if (url is not None) and (text is not None):
            raise PyStockCrawlerError("両方に値を設定しないでください")
        if text is None:
            raise PyStockCrawlerError("textに値が設定されていません")
        result = self.web_scraping(text)
        return result

    def web_scraping(self, text: str) -> dict:
        """
        textより情報を抽出する
        :params text: webページ
        """
        raise NotImplementedError("please implement your code")

    @staticmethod
    def get_url_text(target_url: str) -> str:
        """
        requestsを使って、webからページを取得し、htmlを返す
        """
        user_agent = UserAgent.get_user_agent_header()
        r = requests.get(
            target_url,
            headers=user_agent)

        if r.status_code != 200:
            raise CrawlPageNotFoundError(url=target_url)

        # 日本語に対応
        r.encoding = r.apparent_encoding
        return r.text

    @staticmethod
    def get_crawl_datetime() -> str:
        jst = timezone(timedelta(hours=+9), 'JST')
        now = datetime.now(jst)
        return now.strftime("%Y-%m-%dT%H:%M:%S")



import pandas as pd
from kabutobashi.method.method import Method
from kabutobashi.attributes.attribute import Field


class Momentum(Method):
    """
    https://www.sevendata.co.jp/shihyou/technical/momentum.html
    """
    term = Field()

    def __init__(self, term=25):
        super().__init__(method_name="momentum")
        self.term = term

    def method(self, _df: pd.DataFrame) -> pd.DataFrame:
        _df = _df.assign(
            momentum=_df['close'].shift(10),
            sma_momentum=lambda x: x['momentum'].rolling(self.term).mean()
        )
        return _df

    def signal(self, _df: pd.DataFrame) -> pd.DataFrame:
        _df = _df.join(self.cross(_df['sma_momentum']))
        _df = _df.rename(columns={
            "to_plus": "buy_signal",
            "to_minus": "sell_signal"
        })
        return _df



import sys
import os

# テスト対象のファイルへのパスを通している
# pytestの設定
PARENT_PATH = os.path.abspath(os.path.dirname(os.path.abspath(__file__)))
SOURCE_PATH = PARENT_PATH.rsplit('/', 1)[0]
print(f"test_target_path:= {SOURCE_PATH}")
sys.path.append(f"{SOURCE_PATH}")

# テスト対象のディレクトリに移動
os.chdir("./kabutobashi")


from kabutobashi.crawler.crawler import Crawler
from kabutobashi.crawler.stock_crawler.weeks_52_high_low_page import (
    Week52HighLowStockPricePageTable
)
from bs4 import BeautifulSoup


# 52週の高値・底値を取得する関数とURL
BASE_URL = "https://jp.tradingview.com/markets/stocks-japan"
WEEK_52_HIGH_PRICE_URL = f"{BASE_URL}/highs-and-lows-52wk-high/"
WEEK_52_LOW_PRICE_URL = f"{BASE_URL}/highs-and-lows-52wk-low/"
NEWLY_HIGH_PRICE_UPDATED = f"{BASE_URL}/highs-and-lows-ath/"
NEWLY_LOW_PRICE_UPDATED = f"{BASE_URL}/highs-and-lows-atl/"


def get_52_weeks_high_low(crawl_objective: str) -> dict:
    target_url = None
    if crawl_objective == "high":
        target_url = WEEK_52_HIGH_PRICE_URL
    elif crawl_objective == "low":
        target_url = WEEK_52_LOW_PRICE_URL
    elif crawl_objective == "newly_high":
        target_url = NEWLY_HIGH_PRICE_UPDATED
    elif crawl_objective == "newly_low":
        target_url = NEWLY_LOW_PRICE_UPDATED
    else:
        raise Exception

    # instanceから直接__call__を呼び出して結果を出力
    crawler = Week52HighLowStockPriceCrawler()
    return crawler(url=target_url)


class Week52HighLowStockPriceCrawler(Crawler):
    """
    52週の高値・底値を取得するCrawler
    """

    def __init__(self):
        """
        :params crawl_objective: 取得対象を表す。
        """
        super().__init__()

    def web_scraping(self, text: str) -> dict:
        res = BeautifulSoup(text, 'lxml')

        content = res.find('tbody', class_="tv-data-table__tbody")
        page = Week52HighLowStockPricePageTable(content)
        return page.get_info()



from typing import Union, Optional
import pandas as pd
from datetime import datetime


def example_data() -> pd.DataFrame:
    data_path_list = ["../data/stooq.csv"]
    _df = read_csv(data_path_list)
    _df = _df.sort_values("date", ascending=True)
    _df = _df.convert_dtypes()
    return _df


def read_csv(path_candidate: Union[str, list]) -> Optional[pd.DataFrame]:
    """
    通常のread_csvの関数に加えて、strとlist[str]の場合に縦方向に結合してDataFrameを返す
    :param path_candidate: "path" or ["path_1", "path_2"]
    :return:
    """
    if type(path_candidate) is str:
        return pd.read_csv(path_candidate)
    elif type(path_candidate) is list:
        if not path_candidate:
            return None
        df_list = [pd.read_csv(p) for p in path_candidate]
        return pd.concat(df_list)
    else:
        return None


def read_stock_csv(path_candidate: Union[str, list]) -> Optional[pd.DataFrame]:
    """
    本APIにてCrawlしたデータを扱いやすい形式にデータ変換する関数
    """
    df = read_csv(path_candidate)
    if df is None:
        return None
    else:
        return _decode_stock_data(_df=df)


def _decode_stock_data(_df: pd.DataFrame) -> pd.DataFrame:
    """
    以下のような株のデータを扱いやすいように整形する関数
    stock_label,name,close,date,industry_type,open,high,low,unit,per,psr,pbr,volume,market_capitalization,issued_shares,crawl_datetime,code
    1436  東証マザーズ,フィット,540.0,株価(15:00),業種建設業,555.0円,557.0円,482.0円,100株,---,0.46倍,0.54倍,"22,000株","2,312百万円","4,282千株",2020-03-13T23:31:04,1436
    1438  名証２部,岐阜造園,"1,098.0",株価(12:38),業種建設業,"1,110.0円","1,150.0円","1,098.0円",100株,9.19倍,0.38倍,0.62倍,"3,100株","1,594百万円","1,451千株",2020-03-13T23:31:06,1438
    :param _df:
    :return:
    """

    # 正規表現を利用して数値のみにする
    _df = _df.assign(
        market=_df['stock_label'].str.extract('[0-9]+ (.+)', expand=False),
        open=_df['open'].str.extract('(.+)円', expand=False),
        high=_df['high'].str.extract('(.+)円', expand=False),
        low=_df['low'].str.extract('(.+)円', expand=False),
        unit=_df['unit'].str.extract('(.+)株', expand=False),
        per=_df['per'].str.extract('(.+)倍', expand=False),
        psr=_df['psr'].str.extract('(.+)倍', expand=False),
        pbr=_df['pbr'].str.extract('(.+)倍', expand=False),
        volume=_df['volume'].str.extract('(.+)株', expand=False),
        dt=_df['crawl_datetime'].apply(lambda x: datetime.strptime(x, "%Y-%m-%dT%H:%M:%S").strftime("%Y-%m-%d"))
    )

    # 必要なカラムに絞る
    required_columns = ["code", "open", "close", "high", "low", "unit", "volume", "per", "psr", "pbr", "market", "dt"]
    _df = _df.loc[:, required_columns].drop_duplicates()
    return _df



from kabutobashi.errors import TagNotFoundError


class PageContent(object):
    """
    WebPageからCrawlしてきた値を保持するクラス。
    値を入力する際に改行や空白などの文字を自動で削除する
    """
    def __init__(
            self,
            tag1: str = None,
            _class1: str = None,
            _id1: str = None,
            tag2: str = None,
            _class2: str = None,
            _id2: str = None,
            required=False,
            alternative_data=None,
            required_type=None,
            value_candidate=None):
        """
        :params required: True if value is required
        :params required_type: specify type if the type is fixed
        :params value_candidte: list the candidate values
        :params find_all: find()ではなく、find_all()でデータを取得する
        """
        self.name = None
        self.internal_name = None
        # tagを取得する方法
        self.tag1 = tag1
        self._class1 = {"class": _class1}
        self._id1 = {"id": _id1}
        self.tag2 = tag2
        self._class2 = {"class": _class2}
        self._id2 = {"id": _id2}
        # 値を設定する際の条件など
        self.required: bool = required
        self.alternative_data = alternative_data
        self.required_type: type = required_type
        self.value_candidate: list = value_candidate

    def __get__(self, instance, instance_type):
        return getattr(instance, self.internal_name, None)

    def __set__(self, instance, value):
        if value is None:
            raise ValueError(f"The field is required and none is invalid")

        set_value = None
        # tag1から取得
        if self.tag1 is not None:
            if self._class1['class'] is not None:
                set_value = value.find(self.tag1, self._class1)
            else:
                set_value = value.find(self.tag1)

        # 値がない場合はerror
        if set_value is None and self.required:
            raise TagNotFoundError(tag=self.tag1)

        # tag2もある場合は、追加で取得
        if self.tag2 is not None:
            if self._class2['class'] is not None:
                set_value = set_value.find(self.tag2, self._class2)
            else:
                set_value = set_value.find(self.tag2)

        # 値がない場合はerror
        if set_value is None and self.required:
            raise TagNotFoundError(tag=self.tag2)

        # 文字列を置換して保持
        if set_value is not None:
            set_value = self.replace(set_value.get_text())
        else:
            set_value = self.alternative_data
        setattr(instance, self.internal_name, set_value)

    @staticmethod
    def replace(input_text: str) -> str:
        return input_text.replace(" ", "") \
            .replace("\t", "").replace("\n", "").replace("\r", "") \
            .replace("\xa0", " ").replace("円", "")



import pandas as pd
from kabutobashi.method.api import Method
from typing import Union
# http://www.kabuciao.com/tech/deki/


def analysis_with(
        stock_df: pd.DataFrame,
        method: Method) -> pd.DataFrame:
    """
    :params stock_df:
    :params method:
    """
    return stock_df.pipe(method)


def get_impact_with(
        stock_df: pd.DataFrame,
        method: Union[Method, list],
        **kwargs) -> dict:
    """
    :params stock_df:
    :params method:
    :return: ex: {
            "sma": 0.4,
            "macd": -0.04
        }
    """
    # methodのpipeで渡す際の引数、impactにtrueを渡して直近の各手法の買い・売りの傾向を取得する
    kwargs.update({"impact": "true"})

    # 分析のリスト
    method_list = []
    if type(method) is list:
        method_list.extend(method)
    else:
        method_list.append(method)

    # 結果を格納するdict
    return {str(m): stock_df.pipe(m, **kwargs) for m in method_list}



import pandas as pd
from kabutobashi.errors import StockDfError


class StockDf(object):
    """
    様々な値の保持に用いるクラス。
    値を代入する際にバリデーションが行われる。
    """
    def __init__(self):
        """
        """
        self.name = None
        self.internal_name = None

    def __get__(self, instance, instance_type):
        return getattr(instance, self.internal_name, None)

    def __set__(self, instance, value):
        if value is None:
            raise StockDfError("required")

        df_columns = value.columns
        if "code" in df_columns:
            code = list(set(value.code.values))
            if len(code) > 1:
                raise StockDfError("multiple code")
            elif len(code) == 0:
                raise StockDfError("no code")

        # 日付カラムの候補値を探す
        date_column = None
        if "date" in df_columns:
            date_column = "date"
        elif "dt" in df_columns:
            date_column = "dt"
        if date_column is None:
            raise StockDfError("日付のカラム[dt, date]のいずれかが存在しません")
        if "date" in df_columns and "dt" in df_columns:
            raise StockDfError("日付のカラム[dt, date]は片方しか存在できません")

        # indexにdateを指定
        value.index = pd.to_datetime(value[date_column])

        # 必要なカラムに絞る
        value = value.loc[:, ["open", "high", "low", "close"]]
        open_s = value['open'].apply(self._replace_comma)
        close_s = value['close'].apply(self._replace_comma)
        high_s = value['high'].apply(self._replace_comma)
        low_s = value['low'].apply(self._replace_comma)
        new_value = pd.DataFrame({"open": open_s, "high": high_s, "low": low_s, "close": close_s})
        # 型の指定
        setattr(instance, self.internal_name, new_value)

    @staticmethod
    def _replace_comma(x) -> float:
        """
        pandas内の値がカンマ付きの場合に、カンマを削除する関数
        :param x:
        :return:
        """
        if type(x) is str:
            x = x.replace(",", "")
        try:
            f = float(x)
        except ValueError as e:
            raise StockDfError(f"floatに変換できる値ではありません。{e}")
        return f



import pandas as pd
from kabutobashi.method.method import Method
from kabutobashi.attributes.attribute import Field


class SMA(Method):

    short_term = Field(required_type=int)
    medium_term = Field(required_type=int)
    long_term = Field(required_type=int)

    def __init__(
            self,
            short_term: int = 5,
            medium_term: int = 21,
            long_term: int = 70):
        super().__init__(method_name="sma")
        self.short_term = short_term
        self.medium_term = medium_term
        self.long_term = long_term

    def method(self, _df: pd.DataFrame) -> pd.DataFrame:
        _df = _df.assign(
            sma_short=_df['close'].rolling(self.short_term).mean(),
            sma_medium=_df['close'].rolling(self.medium_term).mean(),
            sma_long=_df['close'].rolling(self.long_term).mean()
        )
        return _df

    def signal(self, _df: pd.DataFrame) -> pd.DataFrame:
        _df['diff'] = _df.apply(lambda x: x['sma_long'] - x['sma_short'], axis=1)
        # 正負が交差した点
        _df = _df.join(self.cross(_df['diff']))
        _df = _df.rename(columns={
            "to_plus": "buy_signal",
            "to_minus": "sell_signal"
        })
        return _df

    # def visualize(self, _df: pd.DataFrame):
    #     fig, ax = plt.subplots(nrows=1, ncols=1, figsize=(6, 5))
    #     # x軸のオートフォーマット
    #     fig.autofmt_xdate()
    #
    #     # set candlestick
    #     self.add_ax_candlestick(ax, _df)
    #
    #     # plot macd
    #     ax.plot(_df.index, _df['sma_long'], color="#dc143c", label="sma_long")
    #     ax.plot(_df.index, _df['sma_medium'], color="#ffa500", label="sma_medium")
    #     ax.plot(_df.index, _df['sma_short'], color="#1e90ff", label="sma_short")
    #
    #     ax.legend(loc="best")  # 各線のラベルを表示
    #     return fig
